/*
 * mod_ntlm.c: NTLM authentication module for Apache/Unix
 * Version 0.1
 * 
 *     "This product includes software developed by the Apache Group
 *     for use in the Apache HTTP server project (http://www.apache.org/)."
 * 
 * Based on 
 * mod_ntlm.c for Win32 by Tim Costello <tim.costello@bigfoot.com>
 * pam_smb by Dave Airlie <Dave.Airlie@ul.ie>
 *
 * This code is copyright 2000 Andreas Gal <agal@uwsp.edu>.
 * Visit http://modntlm.sourceforge.net/ for code updates.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS`` AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES ARE DISCLAIMED. 
 * 
 * This code may be freely distributed, as long the above notices are
 * reproduced.
 */

#define VERSION "0.1"

#include "mod_ntlm.h"

/*#define DEBUG(x) ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, x " %u %u", r->connection, getpid())*/
#define DEBUG(x)

static const command_rec ntlm_cmds[] =
{
    {"NTLMAuth", ap_set_flag_slot,
     (void *) XtOffsetOf(ntlm_config_rec, ntlm_on), OR_AUTHCFG, FLAG,
     "set to 'on' to activate NTLM authentication here"},
    {"NTLMAuthoritative", ap_set_flag_slot,
     (void *) XtOffsetOf(ntlm_config_rec, ntlm_authoritative), OR_AUTHCFG, FLAG,
     "set to 'off' to allow access control to be passed along to lower modules "
     "if the UserID is not known to this module"},
    {"NTLMDomain", ap_set_string_slot,
     (void *) XtOffsetOf(ntlm_config_rec, ntlm_domain), OR_AUTHCFG, TAKE1,
     "set to the domain you want users authenticated against for cleartext "
     "authentication - if not specified, the local machine, then all trusted "
     " domains are checked"},
    {"NTLMServer", ap_set_string_slot,
     (void *) XtOffsetOf(ntlm_config_rec, ntlm_server), OR_AUTHCFG, TAKE1,
     "set to the NT server to contact to authenticate users"},
    {"NTLMBackup", ap_set_string_slot,
     (void *) XtOffsetOf(ntlm_config_rec, ntlm_backup), OR_AUTHCFG, TAKE1,
     "set to the alternate NT server to contact to authenticate users"},
    {NULL}
};

ntlm_connection_rec *ntlm_connection = NULL;

static void *create_ntlm_dir_config(pool *p, char *d)
{
    ntlm_config_rec *crec = (ntlm_config_rec *) ap_pcalloc(p, sizeof(ntlm_config_rec));

    /* Set the defaults. */
    crec->ntlm_authoritative = 1;
    crec->ntlm_on = 0;
    crec->ntlm_server = "SERVER";
    crec->ntlm_backup = "";
    crec->ntlm_domain = "DOMAIN";

    return crec;
}

/* UUENCODE / DECODE TABLES */

static const unsigned char pr2six[256] =
{
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64, 64, 64, 63, 52, 53, 54,
    55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64, 64, 0, 1, 2, 3,
    4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21,
    22, 23, 24, 25, 64, 64, 64, 64, 64, 64, 26, 27, 28, 29, 30, 31, 32,
    33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
    50, 51, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64
};

static const char basis_64[] = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/* 
 * UUENCODE / DECODE ROUTINES BELOW TAKEN FROM APACHE SOURCE CODE
 */

static void *uudecode_binary(pool *p, const char *bufcoded, int *nbytesdecoded)
{
    register const unsigned char *bufin;
    register char *bufplain;
    register unsigned char *bufout;
    register int nprbytes;

    /* Strip leading whitespace. */

    while (*bufcoded == ' ' || *bufcoded == '\t')
    bufcoded++;

    /* Figure out how many characters are in the input buffer.
     * Allocate this many from the per-transaction pool for the result.
     */
#ifndef CHARSET_EBCDIC
    bufin = (const unsigned char *) bufcoded;
    while (pr2six[*(bufin++)] <= 63);
    nprbytes = (bufin - (const unsigned char *) bufcoded) - 1;
    *nbytesdecoded = ((nprbytes + 3) / 4) * 3;

    bufplain = ap_palloc(p, *nbytesdecoded + 1);
    bufout = (unsigned char *) bufplain;

    bufin = (const unsigned char *) bufcoded;

    while (nprbytes > 0) {
    *(bufout++) =
        (unsigned char) (pr2six[*bufin] << 2 | pr2six[bufin[1]] >> 4);
    *(bufout++) =
        (unsigned char) (pr2six[bufin[1]] << 4 | pr2six[bufin[2]] >> 2);
    *(bufout++) =
        (unsigned char) (pr2six[bufin[2]] << 6 | pr2six[bufin[3]]);
    bufin += 4;
    nprbytes -= 4;
    }

    if (nprbytes & 03) {
    if (pr2six[bufin[-2]] > 63)
        *nbytesdecoded -= 2;
    else
        *nbytesdecoded -= 1;
    }
    bufplain[*nbytesdecoded] = '\0';
#else /*CHARSET_EBCDIC*/
    bufin = (const unsigned char *) bufcoded;
    while (pr2six[os_toascii[(unsigned char)*(bufin++)]] <= 63);
    nprbytes = (bufin - (const unsigned char *) bufcoded) - 1;
    *nbytesdecoded = ((nprbytes + 3) / 4) * 3;

    bufplain = ap_palloc(p, *nbytesdecoded + 1);
    bufout = (unsigned char *) bufplain;

    bufin = (const unsigned char *) bufcoded;

    while (nprbytes > 0) {
    *(bufout++) = os_toebcdic[
        (unsigned char) (pr2six[os_toascii[*bufin]] << 2 | pr2six[os_toascii[bufin[1]]] >> 4)];
    *(bufout++) = os_toebcdic[
        (unsigned char) (pr2six[os_toascii[bufin[1]]] << 4 | pr2six[os_toascii[bufin[2]]] >> 2)];
    *(bufout++) = os_toebcdic[
        (unsigned char) (pr2six[os_toascii[bufin[2]]] << 6 | pr2six[os_toascii[bufin[3]]])];
    bufin += 4;
    nprbytes -= 4;
    }

    if (nprbytes & 03) {
    if (pr2six[os_toascii[bufin[-2]]] > 63)
        *nbytesdecoded -= 2;
    else
        *nbytesdecoded -= 1;
    }
    bufplain[*nbytesdecoded] = '\0';
#endif /*CHARSET_EBCDIC*/
    return bufplain;
}

static char *uuencode_binary(pool *a, unsigned char *string, int len) 
{ 
    int i;
    char *p; 
    char *encoded = (char *) ap_palloc(a, ((len+2) / 3 * 4) + 1);
 
    p = encoded; 
#ifndef CHARSET_EBCDIC
    for (i = 0; i < len-2; i += 3) { 
        *p++ = basis_64[(string[i] >> 2) & 0x3F]; 
        *p++ = basis_64[((string[i] & 0x3) << 4) | ((int) (string[i + 1] & 0xF0) >> 4)]; 
        *p++ = basis_64[((string[i + 1] & 0xF) << 2) | ((int) (string[i + 2] & 0xC0) >> 6)]; 
        *p++ = basis_64[string[i + 2] & 0x3F]; 
    } 
    if (i < len) {
        *p++ = basis_64[(string[i] >> 2) & 0x3F]; 
       *p++ = basis_64[((string[i] & 0x3) << 4) | ((int) (string[i + 1] & 0xF0) >> 4)]; 
       if (i == (len-2))
           *p++ = basis_64[((string[i + 1] & 0xF) << 2)]; 
       else
           *p++ = '='; 
       *p++ = '='; 
    }
#else /*CHARSET_EBCDIC*/
    for (i = 0; i < len-2; i += 3) { 
        *p++ = basis_64[(os_toascii[string[i]] >> 2) & 0x3F]; 
        *p++ = basis_64[((os_toascii[string[i]] & 0x3) << 4) | ((int) (os_toascii[string[i + 1]] & 0xF0) >> 4)]; 
        *p++ = basis_64[((os_toascii[string[i + 1]] & 0xF) << 2) | ((int) (os_toascii[string[i + 2]] & 0xC0) >> 6)]; 
        *p++ = basis_64[os_toascii[string[i + 2]] & 0x3F]; 
    } 
    if (i < len) {
        *p++ = basis_64[(os_toascii[string[i]] >> 2) & 0x3F]; 
       *p++ = basis_64[((os_toascii[string[i]] & 0x3) << 4) | ((int) (os_toascii[string[i + 1]] & 0xF0) >> 4)]; 
       if (i == (len-2))
           *p++ = basis_64[((os_toascii[string[i + 1]] & 0xF) << 2)]; 
       else
           *p++ = '='; 
       *p++ = '='; 
    }
#endif /*CHARSET_EBCDIC*/

    *p = '\0'; 
    return encoded; 
} 

static void cleanup_ntlm_connection(void *unused)
{
    if (ntlm_connection->handle) {
        NTLM_Disconnect(ntlm_connection->handle);
        ntlm_connection->handle = NULL;
    }
    ntlm_connection = NULL; /* will be freed with r->connection's context */
}

static void note_ntlm_auth_failure(request_rec *r)
{
    unsigned char *line;

    line = ap_pstrdup(r->pool, NTLM_AUTH_NAME);

    ap_table_setn(r->err_headers_out,
            r->proxyreq ? "Proxy-Authenticate" : "WWW-Authenticate",
            line);
}

static void log_ntlm_logon_denied(request_rec *r)
{
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
        "user %s: authentication failure for \"%s\"",
        ntlm_connection->user, r->uri);
}

ntlmssp_info_rec *get_ntlm_header(request_rec *r, ntlm_config_rec *crec)
{
    const char *auth_line = ap_table_get(r->headers_in,
                                    r->proxyreq ? "Proxy-Authorization"
                                    : "Authorization");
    unsigned char *msg;
    int len;
    ntlmssp_info_rec *ntlmssp;

    if (!auth_line)
        return NULL;

    if (strcmp(ap_getword_white(r->pool, &auth_line), NTLM_AUTH_NAME))
        return NULL;

    msg = uudecode_binary(r->connection->pool, auth_line, &len);
    ntlmssp = ap_pcalloc(r->pool, sizeof(ntlmssp_info_rec));
    if (ntlm_decode_msg(ntlmssp, msg, len) != 0)
        return NULL;

    return ntlmssp;
}

static int send_ntlm_challenge(request_rec *r, ntlm_config_rec *crec)
{
    struct ntlm_msg2 msg;
    unsigned char *challenge;

    DEBUG("received msg1");

    if (ntlm_connection->handle == NULL) {
        ntlm_connection->nonce = ap_pcalloc(r->connection->pool, NONCE_LEN);
        ntlm_connection->handle = NTLM_Connect(crec->ntlm_server,
            crec->ntlm_backup, crec->ntlm_domain, ntlm_connection->nonce);

        if (!ntlm_connection->handle)
            return HTTP_INTERNAL_SERVER_ERROR;
    }

    ntlm_encode_msg2(ntlm_connection->nonce, &msg);
    challenge = uuencode_binary(r->pool, (unsigned char*)&msg, sizeof(msg));

    ap_table_setn(r->err_headers_out,
        r->proxyreq ? "Proxy-Authenticate" : "WWW-Authenticate",
        ap_psprintf(r->pool, "%s %s", NTLM_AUTH_NAME, challenge));

    return HTTP_UNAUTHORIZED;
}

static int ntlm_check_response(request_rec *r, ntlm_config_rec *crec, ntlmssp_info_rec *ntlmssp)
{
    DEBUG("received msg3");

    if (ntlm_connection->user) { /* user has already valid credentials */
        if (   (!strcmp(ntlm_connection->user, ntlmssp->user))
            && (!strcmp(ntlm_connection->domain, ntlmssp->domain))
            && (!memcmp(ntlm_connection->password, ntlmssp->nt, RESP_LEN))) {
            DEBUG("silent reauthentication");
            /* silently accept login with same credentials */
            r->connection->user = ap_pstrdup(r->connection->pool, ntlm_connection->user);
            r->connection->ap_auth_type = ap_pstrdup(r->connection->pool, NTLM_AUTH_NAME);
            return OK;
        }
    }

    if (!ntlm_connection->handle) {
        DEBUG("PDC connection already closed");
        note_ntlm_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    if (!*ntlmssp->user)
        return HTTP_BAD_REQUEST;
    ntlm_connection->user = ap_pstrdup(r->connection->pool, ntlmssp->user);
    ntlm_connection->domain =   (*ntlmssp->domain)
                              ? ap_pstrdup(r->connection->pool, ntlmssp->domain)
                              : crec->ntlm_domain;
    ntlm_connection->password = ap_pcalloc(r->connection->pool, RESP_LEN);
    memcpy(ntlm_connection->password, ntlmssp->nt, RESP_LEN);

    DEBUG("authenticating user against DC");

    if (NTLM_Auth(ntlm_connection->handle,
                  ntlm_connection->user,
                  ntlm_connection->password, 1) == NTV_LOGON_ERROR) {
        log_ntlm_logon_denied(r);
        note_ntlm_auth_failure(r);
        return HTTP_UNAUTHORIZED;
    }

    /*NTLM_Disconnect(ntlm_connection->handle);
    ntlm_connection->handle = NULL;*/
    r->connection->user = ap_pstrdup(r->connection->pool, ntlm_connection->user);
    r->connection->ap_auth_type = ap_pstrdup(r->connection->pool, NTLM_AUTH_NAME);

    return OK;
}

static int authenticate_ntlm_user(request_rec *r)
{
    ntlm_config_rec *crec =
    (ntlm_config_rec *) ap_get_module_config(r->per_dir_config, &ntlm_module);
    ntlmssp_info_rec *ntlmssp;
    int res;

    if (!crec->ntlm_on)
        return DECLINED;

    /* If this is the first request with this connection, then create a
     * ntlm_connection entry for it. It will be cleaned up when the
     * connection is dropped */
    if (ntlm_connection == NULL) {
        ntlm_connection = ap_pcalloc(r->connection->pool, sizeof(ntlm_connection_rec));
        ap_register_cleanup(r->connection->pool, 0, cleanup_ntlm_connection, ap_null_cleanup);
    }

    if ((ntlmssp = get_ntlm_header(r, crec)) == NULL) {
        note_ntlm_auth_failure(r);
        DEBUG("missing/corrupt NTLM header");
        return HTTP_UNAUTHORIZED;
    }

    switch (ntlmssp->msg_type) {
    case 1:
        return send_ntlm_challenge(r, crec);
    case 3:
        return ntlm_check_response(r, crec, ntlmssp);
    }

    return HTTP_BAD_REQUEST;
}

static int check_user_access(request_rec *r)
{
    ntlm_config_rec *crec =
    (ntlm_config_rec *) ap_get_module_config(r->per_dir_config, &ntlm_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    const array_header *reqs_arr = ap_requires(r);
    require_line *reqs;

    /*
     * If the switch isn't on, don't bother. 
     */
    if (!crec->ntlm_on) {
        return DECLINED;
    }
    
    if (!reqs_arr) {
        return OK;
    }

    reqs = (require_line *) reqs_arr->elts;

    /*
     * Did we authenticate this user?
     * If not, we don't want to do user/group checking.
     */
    if (!ntlm_connection || !ntlm_connection->user)
        return DECLINED;

    for (x = 0; x < reqs_arr->nelts; x++) {
        if (!(reqs[x].method_mask & (1 << m)))
            continue;

        method_restricted = 1;

        t = reqs[x].requirement;
        w = ap_getword_white(r->pool, &t);
        if (!strcmp(w, "valid-user"))
            return OK;
        if (!strcmp(w, "user")) {
            while (t[0]) {
            w = ap_getword_conf(r->pool, &t);
            if (!strcmp(user, w))
                return OK;
            }
        }
        else if (crec->ntlm_authoritative) {
            /* if we aren't authoritative, any require directive could be
             * valid even if we don't grok it.  However, if we are 
             * authoritative, we can warn the user they did something wrong.
             * That something could be a missing "AuthAuthoritative off", but
             * more likely is a typo in the require directive.
             */
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
            "access to %s failed, reason: unknown require directive:"
            "\"%s\"", r->uri, reqs[x].requirement);
        }
    }

    if (!method_restricted) {
        return OK;
    }

    if (!(crec->ntlm_authoritative)) {
        return DECLINED;
    }

    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
        "access to %s failed, reason: user %s not allowed access",
        r->uri, user);
    
    note_ntlm_auth_failure(r);
    /*
     * We return HTTP_UNAUTHORIZED (401) because the client may wish
     * to authenticate using a different scheme, or a different 
     * username. If this works, they can be granted access. If we 
     * returned HTTP_FORBIDDEN (403) then they don't get a second
     * chance. 
     */
    return HTTP_UNAUTHORIZED;
}

module MODULE_VAR_EXPORT ntlm_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                    /* initializer */
    create_ntlm_dir_config,  /* dir config creater */
    NULL,                    /* dir merger --- default is to override */
    NULL,                    /* server config */
    NULL,                    /* merge server config */
    ntlm_cmds,               /* command table */
    NULL,                    /* handlers */
    NULL,                    /* filename translation */
    authenticate_ntlm_user,  /* check_user_id */
    check_user_access,       /* check auth */
    NULL,                    /* check access */
    NULL,                    /* type_checker */
    NULL,                    /* fixups */
    NULL,                    /* logger */
    NULL,                    /* header parser */
    NULL,                    /* child_init */
    NULL,                    /* child_exit */
    NULL                     /* post read-request */
};

